# -*- coding: utf-8 -*-
#===============================================================#
# Oligonucleotide frequency calculation program                 #
# ver. 2020.04                                                  #
# 1st argument: genome seq. (FASTA format)                      #
# 2nd argument: oligonucleotide pattern                         #
# usage: ./freq_calc_ver202004.rb FASTAfile.fas ./ptn/tetra.txt #
#===============================================================#

$:.unshift(File.dirname(__FILE__))

##########################################################

### Parameter ###


# 1. Window and step size for fragmentation
# WINDOW_SIZE :  fragment size 
# STEP_SIZE   : slide size for next fragment
# READ_SIZE   : minimum length that can be calculated
WINDOW_SIZE = 5000
STEP_SIZE = 5000
READ_SIZE = WINDOW_SIZE

# 2. Threshold for the number of undetermined nucleotides
# When the number of undetermined nucleotides in a sequence exceeded 10% of the window size, 
# the sequence was omitted from the calculatiaon.
# 100% : 1.0
#  10% : 0.1
# 0.1% : 0.01
NPAR = 0.1

# 3. Setting the output file extension
end_name ='.frq'

# Calc. type 
# Percent    :per
# Frequency  :freq
# Count      :count
CALC = 'count'


### End of Parameter###

###########################################################




##########(main_method)##########
def origonucleotide_freq(flat_file_name, pattern_file_name, end_name = "")
  print "Input file : #{flat_file_name}\n"
  print "Pattern file: #{pattern_file_name}\n"

 #Init. 
  flat_name = File.basename(flat_file_name)
  dna_seq = ""
  word_len = 0
  dimension = 0
  word_pattern = []
  first_line = ""
  second_line = ""
  same_pattern = []
  

  #Read fasta file
  File.open(flat_file_name){|flat_data|
    flat_data.each{|line_data|
      if line_data =~ /^>/ 
        first_line = line_data.chomp
      else
        dna_seq << line_data.chomp
      end
    }
  }
  
  
  File.open(pattern_file_name){|pattern|
    word_len = pattern.gets.chomp.to_i   
    dimension = pattern.gets.chomp.to_i  

    pattern.each{|line|
      word_pattern.push(line.chomp)      


      if line =~ /\+/
        column = line.split(/\+/)      
        pattern_left = column[0]
        pattern_right = column[1].chomp
        
        if pattern_left == pattern_right
          same_pattern.push(pattern_left) 
        end 
      end
    }
  }


  dna_seq.upcase!  
   
  dna_seq_len = dna_seq.length   
  
  
  #======================================
  if dna_seq_len >= READ_SIZE
  #======================================
    
    
    #塩基配列を断片化させる
    dna_seq_frag_ary = fragmentation(dna_seq, dna_seq_len)
    print "Total fragment number : #{dna_seq_frag_ary.length}\n"

    i = 0
    sp = 0
    
    #calc. each fragment    
    name = flat_file_name.chomp + end_name
    w_file = File.open("#{name}", "w")

    dna_seq_frag_ary.each do |dna_seq_ary|
      
      #Check Ns
      if contentNunder10(dna_seq_ary)
        
        origonucleotide_freq = calculate_origonucleotide_freq(dna_seq_ary, word_len, same_pattern)
        
        #comment line
        if (sp + WINDOW_SIZE) > dna_seq_len          
          comment = i.to_s + " [" + (dna_seq_len - WINDOW_SIZE ).to_s + "-" + (dna_seq_len - 1).to_s + "] < " + flat_name
        else
          comment = i.to_s + " [" + sp.to_s + "-" + (sp + WINDOW_SIZE - 1).to_s + "] < " + flat_name
        end

        i += 1
        sp += STEP_SIZE
        
        #output
        output(flat_name, comment, origonucleotide_freq, word_pattern, w_file)

      else
        sp += STEP_SIZE
      end      
    end
    w_file.close
  else
    print("\nError : seq less windowsize\nPlease check parameter in this program.\n\n") 
  end
end



def fragmentation(dna_seq_str, dna_len = nil)

  dna_seq_frag_ary = []
  
  dna_seq_ary = dna_seq_str.split("")
  
  if !dna_len
    dna_len = dna_seq_str.length
  end
  
  sp = 0
  ep = sp + WINDOW_SIZE - 1  
  i = 0
  
  while sp < dna_len
    
    if ep > dna_len - 1
      sp = dna_len - WINDOW_SIZE
      ep = dna_len -1
      dna_seq_frag_ary[i] = dna_seq_ary[sp..ep]
    else
      dna_seq_frag_ary[i] = dna_seq_ary[sp..ep]
    end
    
    sp += STEP_SIZE
    ep += STEP_SIZE
    i += 1

  end
  
  return dna_seq_frag_ary
  
end



########## Check Ns ##########
def contentNover10(dna_seq_ary)
  
  count = 0
  dna_len = dna_seq_ary.length
  
  dna_seq_ary.each do |nc|
    if nc !~ /A|a|T|t|G|g|C|c/
      count += 1
    end
  end
  
  #======================================================================
  if (count.to_f / dna_len.to_f) >= NPAR
  #======================================================================
    return true
  else
    return false
  end

end

def contentNunder10(dna_seq_ary)
  !(contentNover10(dna_seq_ary))
end



def calculate_origonucleotide_freq(dna_seq_ary, word_len, pattern)
  onc = origonucleotide_count(dna_seq_ary, word_len)
  dna_len = dna_seq_ary.length
  
  word_total = dna_len - (word_len - 1)

  onc_freq_ConsiderlateSingleNC = Hash.new(0)
  
  onc.each do |word, count|
    if word !~ /[A|a|T|t|G|g|C|c]{#{word_len}}/
      word_total -= count
      onc.delete(word)
    end
  end
  
  
  onc.each do |word, count|
    i = 0
    while line = pattern[i]
      i += 1
      if word == line
        word_total += count
      end
    end
  end
  
  
  onc.each{|word, count,|

    #Clac.
    case CALC      
    when "per"   #percent

      onc_freq = count.to_f / word_total
      onc_freq_ConsiderlateSingleNC[word] = onc_freq * 100
            
    when "frq"   #frequency

      onc_freq = count.to_f / word_total
      onc_freq_ConsiderlateSingleNC[word] = onc_freq

    when "count"   #count

      onc_freq = count.to_f
      onc_freq_ConsiderlateSingleNC[word] = onc_freq

    end


  }
  
  return onc_freq_ConsiderlateSingleNC
end



def origonucleotide_count(dna_seq_ary, word_len)

  word = ""
  onc_count = Hash.new(0)
  i = 0

  while dna_seq_ary[i + word_len - 1]
    word_len.times do |j|
      word << dna_seq_ary[i+j]
    end
    
    onc_count[word] += 1
    word = ""
    i += 1
  end
  
  return onc_count
end



def output(first_line, second_line, freq_data, word_pattern, w_file)
  w_file.print ">#{first_line}\n"
#  w_file.print "#{first_line}\n"
  w_file.print "#{second_line}\n"
  
  word_pattern.each{|ptn|
    if ptn =~ /\+/
      tmp = ptn.split('+')
      tmp[1].chomp
      w_file.print(freq_data[tmp[0]] + freq_data[tmp[1]], ' ')
    else
      w_file.print(freq_data[ptn], ' ')
    end
  }
  w_file.print("\n")
end



###########(main)############
pattern_file_name = ARGV[1]

ARGV[0].each{|input_file|  
  origonucleotide_freq(input_file, pattern_file_name, end_name)
  print "Program End.\n"
}
